# make_stub_files: Fri 21 Jun 2019 at 00:44:00

from typing import Any, Dict, Optional, Sequence, Tuple, Union

Node = Any
def time() -> Any: ...
    #   0: return val-NTP_DELTA
    # ? 0: return val-NTP_DELTA
def settime() -> None: ...
