"""
Module: 'machine' on esp8266 v1.9.4
"""
# MCU: (sysname='esp8266', nodename='esp8266', release='2.2.0-dev(9422289)', version='v1.9.4-8-ga9a3caad0 on 2018-05-11', machine='ESP module with ESP8266')
# Stubber: 1.1.2


class ADC:
    ""

    def read():
        pass


DEEPSLEEP = 4
DEEPSLEEP_RESET = 5
HARD_RESET = 6


class I2C:
    ""
